"""
Useful geometric operations, e.g. Orthographic projection and a differentiable Rodrigues formula

Parts of the code are taken from https://github.com/MandyMo/pytorch_HMR
"""

# copy from https://github.com/microsoft/MeshTransformer/blob/main/metro/utils/geometric_layers.py

import torch
from torch.nn import functional as F

def rodrigues(theta):
    """Convert axis-angle representation to rotation matrix.
    Args:
        theta: size = [B, 3]
    Returns:
        Rotation matrix corresponding to the quaternion -- size = [B, 3, 3]
    """
    l1norm = torch.norm(theta + 1e-8, p = 2, dim = 1)
    angle = torch.unsqueeze(l1norm, -1)
    normalized = torch.div(theta, angle)
    angle = angle * 0.5
    v_cos = torch.cos(angle)
    v_sin = torch.sin(angle)
    quat = torch.cat([v_cos, v_sin * normalized], dim = 1)
    return quat2mat(quat)

def quat2mat(quat):
    """Convert quaternion coefficients to rotation matrix.
    Args:
        quat: size = [B, 4] 4 <===>(w, x, y, z)
    Returns:
        Rotation matrix corresponding to the quaternion -- size = [B, 3, 3]
    """ 
    norm_quat = quat
    norm_quat = norm_quat/norm_quat.norm(p=2, dim=1, keepdim=True)
    w, x, y, z = norm_quat[:,0], norm_quat[:,1], norm_quat[:,2], norm_quat[:,3]

    B = quat.size(0)

    w2, x2, y2, z2 = w.pow(2), x.pow(2), y.pow(2), z.pow(2)
    wx, wy, wz = w*x, w*y, w*z
    xy, xz, yz = x*y, x*z, y*z

    rotMat = torch.stack([w2 + x2 - y2 - z2, 2*xy - 2*wz, 2*wy + 2*xz,
                          2*wz + 2*xy, w2 - x2 + y2 - z2, 2*yz - 2*wx,
                          2*xz - 2*wy, 2*wx + 2*yz, w2 - x2 - y2 + z2], dim=1).view(B, 3, 3)
    return rotMat    
    
def orthographic_projection(X, camera):
    """Perform orthographic projection of 3D points X using the camera parameters
    Args:
        X: size = [B, N, 3]
        camera: size = [B, 3]
    Returns:
        Projected 2D points -- size = [B, N, 2]
    """ 
    camera = camera.view(-1, 1, 3)
    X_trans = X[:, :, :2] + camera[:, :, 1:]
    shape = X_trans.shape
    X_2d = (camera[:, :, 0] * X_trans.view(shape[0], -1)).view(shape)
    return X_2d

def perspective_projection(points, translation,
                           focal_length, rotation=None, camera_center=None):
    """
    This function computes the perspective projection of a set of points.
    Input:
        points (bs, N, 3): 3D points
        rotation (bs, 3, 3): Camera rotation
        translation (bs, 3): Camera translation
        focal_length (bs,) or scalar: Focal length
        camera_center (bs, 2): Camera center
    """
    batch_size = points.shape[0]
    K = torch.zeros([batch_size, 3, 3], device=points.device)
    K[:,0,0] = focal_length
    K[:,1,1] = focal_length
    K[:,2,2] = 1.
    if camera_center is not None:
        K[:,:-1, -1] = camera_center

    # Transform points
    if rotation is not None:
        points = torch.einsum('bij,bkj->bki', rotation, points)
    points = points + translation.unsqueeze(1)

    # Apply perspective distortion
    projected_points = points / points[:,:,-1].unsqueeze(-1)

    # Apply camera intrinsics
    projected_points = torch.einsum('bij,bkj->bki', K, projected_points)

    return projected_points[:, :, :-1]

def rot6d_to_rotmat(x):
    """Convert 6D rotation representation to 3x3 rotation matrix.
    Based on Zhou et al., "On the Continuity of Rotation Representations in Neural Networks", CVPR 2019
    Input:
        (B,6) Batch of 6-D rotation representations
    Output:
        (B,3,3) Batch of corresponding rotation matrices
    """
    x = x.view(-1,3,2)
    a1 = x[:, :, 0]
    a2 = x[:, :, 1]
    b1 = F.normalize(a1)
    b2 = F.normalize(a2 - torch.einsum('bi,bi->b', b1, a2).unsqueeze(-1) * b1)
    b3 = torch.cross(b1, b2)
    return torch.stack((b1, b2, b3), dim=-1)





import os

# os.environ['PYOPENGL_PLATFORM'] = 'osmesa'
import torch
from torchvision.utils import make_grid
import numpy as np
import pyrender
import trimesh


class Renderer(object):
    """
    Renderer used for visualizing the SMPL model
    Code adapted from https://github.com/vchoutas/smplify-x
    """

    def __init__(
        self, faces, img_res=224, width=None, height=None, focal_length=5000
    ):
        if width is None:
            width = img_res
        if height is None:
            height = img_res
        self.renderer = pyrender.OffscreenRenderer(
            viewport_width=width, viewport_height=height, point_size=1.0
        )
        self.focal_length = focal_length
        self.camera_center = [img_res // 2, img_res // 2]
        self.faces = faces

    def visualize_tb(self, vertices, camera_translation, images):
        vertices = vertices.cpu().numpy()
        camera_translation = camera_translation.cpu().numpy()
        images = images.cpu()
        images_np = np.transpose(images.numpy(), (0, 2, 3, 1))
        rend_imgs = []
        for i in range(vertices.shape[0]):
            rend_img = torch.from_numpy(
                np.transpose(
                    self.__call__(vertices[i], camera_translation[i], images_np[i]),
                    (2, 0, 1),
                )
            ).float()
            rend_imgs.append(images[i])
            rend_imgs.append(rend_img)
        rend_imgs = make_grid(rend_imgs, nrow=2)
        return rend_imgs

    def __call__(self, **kwargs):
        return self.render(**kwargs)

    def render(self, vertices, camera_translation, image, camera_center=None):
        """
        image: 范围在[0,1]
        """
        if camera_center is None:
            camera_center = self.camera_center
        material = pyrender.MetallicRoughnessMaterial(
            metallicFactor=0.2, alphaMode="OPAQUE", baseColorFactor=(0.8, 0.3, 0.3, 1.0)
        )

        camera_translation[0] *= -1.0

        mesh = trimesh.Trimesh(vertices, self.faces)
        rot = trimesh.transformations.rotation_matrix(np.radians(180), [1, 0, 0])
        mesh.apply_transform(rot)
        mesh = pyrender.Mesh.from_trimesh(mesh, material=material)

        scene = pyrender.Scene(ambient_light=(0.5, 0.5, 0.5))
        scene.add(mesh, "mesh")

        camera_pose = np.eye(4)
        camera_pose[:3, 3] = camera_translation
        camera = pyrender.IntrinsicsCamera(
            fx=self.focal_length,
            fy=self.focal_length,
            cx=camera_center[0],
            cy=camera_center[1],
        )
        scene.add(camera, pose=camera_pose)

        light = pyrender.DirectionalLight(color=[1.0, 1.0, 1.0], intensity=1)
        light_pose = np.eye(4)

        light_pose[:3, 3] = np.array([0, -1, 1])
        scene.add(light, pose=light_pose)

        light_pose[:3, 3] = np.array([0, 1, 1])
        scene.add(light, pose=light_pose)

        light_pose[:3, 3] = np.array([1, 1, 2])
        scene.add(light, pose=light_pose)

        color, rend_depth = self.renderer.render(scene, flags=pyrender.RenderFlags.RGBA)
        color = color.astype(np.float32) / 255.0
        valid_mask = (rend_depth > 0)[:, :, None]

        output_img = color[:, :, :3] * valid_mask + (1 - valid_mask) * image
        return output_img



import os.path as osp
from concurrent.futures import as_completed, ThreadPoolExecutor
import signal
from functools import partial
from threading import Event
from typing import Iterable
import requests
from urllib.request import urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from ..tools.logger import Logger


progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


done_event = Event()


def handle_sigint(signum, frame):
    done_event.set()


signal.signal(signal.SIGINT, handle_sigint)


def copy_url(task_id: TaskID, url: str, path: str) -> None:
    """Copy data from a url to a local file."""

    progress.console.log(f"Requesting {url}")
    response = urlopen(url)
    # This will break if the response doesn't contain content length
    progress.update(task_id, total=int(response.info()["Content-length"]))
    with open(path, "wb") as dest_file:
        progress.start_task(task_id)
        for data in iter(partial(response.read, 32768), b""):
            dest_file.write(data)
            progress.update(task_id, advance=len(data))
            if done_event.is_set():
                return
    progress.console.log(f"Downloaded {path}")


def download(urls: Iterable[str], dest_dir: str):
    """同时下载多个文件

    Args:
        urls: download list
        dest_dir: 指定存储文件夹

    Example:
        urls = [
            'https://releases.ubuntu.com/20.04/ubuntu-20.04.2.0-desktop-amd64.iso',
        ]
        download(urls, "./")
        download(urls, "data/")
    """

    if not Logger.check_path(dest_dir):
        return

    with progress:
        with ThreadPoolExecutor(max_workers=4) as pool:
            for url in urls:
                filename = url.split("/")[-1]
                dest_path = osp.join(dest_dir, filename)
                task_id = progress.add_task("download", filename=filename, start=False)
                pool.submit(copy_url, task_id, url, dest_path)


def download_file(url, filename):
    """ 下载单个文件 """
    content = requests.get(url).content
    with open(filename, "wb") as fw:
        fw.write(content)



import json
import cv2
import pickle
import yaml
import scipy.io as scio
import numpy as np
import os
import os.path as osp

__all__ = ["load_yaml", "load_file", "load_json", "save_file", "save_json", "mkdir"]


def load_json(filename):
    with open(filename, "r") as fr:
        data = json.load(fr)
    return data


def load_img(filename):
    return cv2.imread(filename)


def load_pkl(filename):
    return pickle.load(open(filename, "rb"), encoding="latin1")


def load_yaml(filename):
    return yaml.load(open(filename, "r"), Loader=yaml.CLoader)


def load_ply(filename):
    return o3d.io.read_triangle_mesh(filename)


def load_mat(filename):
    return scio.loadmat(filename)


def load_npz(filename):
    """print(data.files)"""
    return np.load(filename, allow_pickle=True)


def load_txt(filename):
    with open(filename, "r") as fr:
        lines = fr.readlines()
    return lines


def load_obj(obj_filename):
    v_list = []
    f_list = []

    with open(obj_filename, "r") as fr:
        for line in fr.readlines():
            segs = line.split(" ")
            if segs[0] == "v":
                x = float(segs[1])
                y = float(segs[2])
                z = float(segs[3])
                v_list.append([x, y, z])

            elif segs[0] == "f":
                a = int(segs[1])
                b = int(segs[2])
                c = int(segs[3])
                f_list.append([a, b, c])
    vertices = np.array(v_list, dtype=np.float)
    faces = np.array(f_list, dtype=np.int)

    if np.min(faces) == 1:
        faces -= 1
    print("face index start from:", np.min(faces))
    return vertices, faces


def load_file(filename):
    """supported filetype:
    @txt, json, pkl, yaml, ply, mat, npz
    @jpg, jpeg, png
    """
    if filename.endswith(".txt"):
        return load_txt(filename)

    if filename.endswith(".json"):
        return load_json(filename)

    if filename.endswith(".jpg"):
        return load_img(filename)
    if filename.endswith(".jpeg"):
        return load_img(filename)
    if filename.endswith(".png"):
        return load_img(filename)

    if filename.endswith(".pkl"):
        return load_pkl(filename)

    if filename.endswith(".yaml"):
        return load_yaml(filename)

    if filename.endswith(".ply"):
        return load_ply(filename)

    if filename.endswith(".mat"):
        return load_mat(filename)

    # print(data.files)
    if filename.endswith(".npz"):
        return load_npz(filename)

    assert "Wrong file type" and False
    return None


def save_file(filename, data):
    """.json .yaml .mat"""
    file_type = filename.split(".")[-1]

    if file_type == "mat":
        scio.savemat(filename, data)

    elif file_type == "json":
        save_json(filename, data)

    elif file_type == "yaml":
        save_yaml(filename, data)

    else:
        print("file_type", file_type)
        assert False


def save_json(filename, data):
    with open(filename, "w", encoding="utf-8") as fw:
        json.dump(data, fw, ensure_ascii=False, indent=4)


def save_yaml(filename, data):
    with open(filename, "w", encoding="utf-8") as fw:
        yaml.dump(data, fw)


def mkdir(path: str) -> None:
    if not osp.exists(path):
        os.makedirs(path, exist_ok=True)



def get_train_tips():
    return [
        {
            "title": "深度学习网络训练技巧汇总",
            "url": "https://zhuanlan.zhihu.com/p/20767428",
        },
        {
            "title": "深度学习调参有哪些技巧？",
            "url": "https://www.zhihu.com/question/25097993",
        },
    ]



import torch
import numpy as np
import torchvision.transforms as transforms

IMG_MEAN = [0.485, 0.456, 0.406]
IMG_STD = [0.229, 0.224, 0.225]


def tensor_to_bgr(img, normalized=True):
    img = img.permute(1, 2, 0).numpy()

    if normalized:
        mean = np.array(IMG_MEAN)
        std = np.array(IMG_STD)

        img = img * std + mean

    img = img[:, :, ::-1].copy() * 255.0
    img = img.astype(np.uint8)
    return img


def bgr_to_tensor(bgr_img, normalized=True):
    rgb_img = bgr_img[:, :, ::-1]
    img = np.transpose(rgb_img.astype(np.float32), (2, 0, 1)) / 255.0
    img = torch.from_numpy(img).float()
    if normalized:
        img = transforms.Normalize(mean=IMG_MEAN, std=IMG_STD)(img)
    return img

def rgb_to_bgr(img):
    return img[:, :, ::-1]



import psutil
import re
import platform
import torch

__all__ = ["get_cpu_info", "get_gpu_info", "get_package_info", "get_memory_info"]


def get_cpu_info():
    return {
        "system": platform.version(),
        "cpu_name": get_cpu_name(),
        "cpu_count": get_cpu_count(),
        "cpu_threads": get_cpu_threads(),
        "cpu_core": get_cpu_core(),
    }


def get_gpu_info():
    return {
        "gpu_name": torch.cuda.get_device_name(0),
        "gpu_num": torch.cuda.device_count(),
    }


def get_memory_info():
    """GB"""
    pc_mem = psutil.virtual_memory()
    gb_factor = 1024.0 ** 3
    return {
        "total": float(pc_mem.total / gb_factor),
        "available": float(pc_mem.available / gb_factor),
        "used": float(pc_mem.used / gb_factor),
    }


def get_package_info():
    return {
        "python": platform.python_version(),
        "pytorch": torch.__version__,
        "cuda": torch.version.cuda,
        "cudnn": torch.backends.cudnn.version(),
    }


def get_cpu_name():
    cpuinfo = read_file("/proc/cpuinfo")
    rep = "model\s+name\s+:\s+(.+)"
    tmp = re.search(rep, cpuinfo, re.I)
    cpu_name = tmp.groups()[0]
    return cpu_name


def get_cpu_count():
    return len(set(re.findall("physical id.+", read_file("/proc/cpuinfo"))))


def get_cpu_threads():
    return psutil.cpu_count()


def get_cpu_core():
    return psutil.cpu_count(logical=False)


def read_file(filename):
    try:
        with open(filename, "r", encoding="utf-8") as file:
            return file.read()
    except:
        pass

    return ""



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="human_cver",
    version="0.1.3",
    author="Wenkai Liu",
    author_email="wenkai_liu@hust.edu.cn",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wkailiu/human_cver",
    project_urls={
        "Bug Tracker": "https://github.com/wkailiu/human_cver/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.yaml", ],
        # And include any *.msg files found in the 'hello' package, too:
        # "hello": ["*.msg"],
    },
    python_requires=">=3.6",
    install_requires=[
        "rich>=11.0",
        "scipy",
        "opencv-python",
        "logzero",
        "pytorch_lightning",
        "omegaconf",
        "psutil",
        "snakeviz",
        "flameprof",
    ],
    entry_points={
        'console_scripts': [
            'train=human_cver:train_main',
            'make_config=human_cver:make_config',
        ]
    }
)


from typing import List
from torch.utils.data import Dataset
from ..tools.logger import Logger


class MixedDataset(Dataset):
    def __init__(self, dataset_list: List[Dataset]) -> Dataset:
        self.dataset_list = dataset_list
        self.lens = [len(dataset) for dataset in dataset_list]
        total = sum(self.lens)
        Logger.info(f"mixed dataset-> total={total}, lens: {self.lens}")

    def __getitem__(self, index):
        k = index
        for i, len in enumerate(self.lens):
            if k < len:
                return self.dataset_list[i][k]
            k -= len

    def __len__(self):
        return sum(self.lens)



"""
Functions for compuing Procrustes alignment and reconstruction error

Parts of the code are adapted from https://github.com/akanazawa/hmr

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from ..tools.logger import Logger


def compute_similarity_transform(S1, S2):
    """Computes a similarity transform (sR, t) that takes
    a set of 3D points S1 (3 x N) closest to a set of 3D points S2,
    where R is an 3x3 rotation matrix, t 3x1 translation, s scale.
    i.e. solves the orthogonal Procrutes problem.
    """
    transposed = False
    if S1.shape[0] != 3 and S1.shape[0] != 2:
        S1 = S1.T
        S2 = S2.T
        transposed = True
    assert S2.shape[1] == S1.shape[1]

    # 1. Remove mean.
    mu1 = S1.mean(axis=1, keepdims=True)
    mu2 = S2.mean(axis=1, keepdims=True)
    X1 = S1 - mu1
    X2 = S2 - mu2

    # 2. Compute variance of X1 used for scale.
    var1 = np.sum(X1 ** 2)

    # 3. The outer product of X1 and X2.
    K = X1.dot(X2.T)

    # 4. Solution that Maximizes trace(R'K) is R=U*V', where U, V are
    # singular vectors of K.
    U, s, Vh = np.linalg.svd(K)
    V = Vh.T
    # Construct Z that fixes the orientation of R to get det(R)=1.
    Z = np.eye(U.shape[0])
    Z[-1, -1] *= np.sign(np.linalg.det(U.dot(V.T)))
    # Construct R.
    R = V.dot(Z.dot(U.T))

    # 5. Recover scale.
    scale = np.trace(R.dot(K)) / var1

    # 6. Recover translation.
    t = mu2 - scale * (R.dot(mu1))

    # 7. Error:
    S1_hat = scale * R.dot(S1) + t

    if transposed:
        S1_hat = S1_hat.T

    return S1_hat


def compute_similarity_transform_batch(S1, S2):
    """Batched version of compute_similarity_transform."""
    S1_hat = np.zeros_like(S1)
    for i in range(S1.shape[0]):
        try:
            simi = compute_similarity_transform(S1[i], S2[i])
        except:
            simi = 0.0
            Logger.error("np.linalg.svd error.  set simi=0")
        else:
            pass
        S1_hat[i] = simi
    return S1_hat


def reconstruction_error(S1, S2, reduction="mean"):
    """Do Procrustes alignment and compute reconstruction error."""
    S1_hat = compute_similarity_transform_batch(S1, S2)
    re = np.sqrt(((S1_hat - S2) ** 2).sum(axis=-1)).mean(axis=-1)
    if reduction == "mean":
        re = re.mean()
    elif reduction == "sum":
        re = re.sum()
    return re


def reconstruction_error_v2(S1, S2, J24_TO_J14, reduction="mean"):
    """Do Procrustes alignment and compute reconstruction error."""
    S1_hat = compute_similarity_transform_batch(S1, S2)
    S1_hat = S1_hat[:, J24_TO_J14, :]
    S2 = S2[:, J24_TO_J14, :]
    re = np.sqrt(((S1_hat - S2) ** 2).sum(axis=-1)).mean(axis=-1)
    if reduction == "mean":
        re = re.mean()
    elif reduction == "sum":
        re = re.sum()
    return re


def get_alignMesh(S1, S2, reduction="mean"):
    """Do Procrustes alignment and compute reconstruction error."""
    S1_hat = compute_similarity_transform_batch(S1, S2)
    re = np.sqrt(((S1_hat - S2) ** 2).sum(axis=-1)).mean(axis=-1)
    if reduction == "mean":
        re = re.mean()
    elif reduction == "sum":
        re = re.sum()
    return re, S1_hat, S2
